/* *********************************************************************
*Copyright (C) 2002 Elliott Mitchell					*
*All rights reserved.							*
*									*
*$Id: input.c,v 3.2 2002/05/23 06:38:13 ehem Exp $			*
*************************************************************************
*CS410-Combinatorial Search Project					*
************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include "common.h"

struct edge {
	node_t b, e;
};


static uint64 getinput(FILE *, struct instance **, struct edge **);
static int processinput(struct instance *, struct edge *, uint64);
static int digestinput(struct instance *);
static int edgecmp(const struct edge *, const struct edge *);
static node_t colonize_island(struct node *chain, struct node *n);


struct instance *doinput(FILE *in)
{
	struct instance *ret;
	struct edge *e;
	uint64 ecnt;

	if(!(ecnt=getinput(in, &ret, &e))) return(NULL);

	ecnt=processinput(ret, e, ecnt);
	free(e);
	if(!ecnt) return(NULL);

	if(!digestinput(ret)) return(NULL);

	return(ret);
}


uint64 getinput(FILE *in, struct instance **p_ret, struct edge **p_edges)
{
	struct instance *ret=NULL;
	uint64 edgesfound=0;
	struct edge *edges=NULL; /*Don't need to init, but avoids a warning*/
	uint64 ecnt=1; /*Prevent premature termination of the loop*/
	uint64 nodes=0;
#ifdef BITS64
#warning "Unfortunatly we have no larger datatype for our edge counter"
#endif
	int ltype;

	do {
		ltype=getc(in);
		while(isspace(ltype)) ltype=getc(in);

		switch(ltype) {
		case 'c':
			while(ltype!='\n') ltype=getchar();
		break;
		case 'p':
			if(nodes) {
				fprintf(stderr, "Error: Multiple problem "
"lines found\n");
				free(edges);
				return(0);
			}

			fscanf(in, " edge %llu %llu\n", &nodes, &ecnt);

			if(nodes>NODEMAX) {
				fprintf(stderr, "Maximum node count exceeded, "
"either change to BITS32/BITS64, or get more\n" "hardware and make the "
"compiler understand unsigned long long long.\n");
				return(0);
			}

			if(ecnt>(EDGEMAX^(EDGEMAX>>1))) {
				fprintf(stderr, "Capacity of internal "
"variables overloaded. Either implement BITS128, or shrink\n" "the instance\n");
				return(0);
			}

			edges=malloc(sizeof(struct edge)*(2*ecnt+1));

			if(!edges) {
				fprintf(stderr, "Unable to allocate %lld bytes "
"for edge loading space.\n", sizeof(struct edge)*2*ecnt);
				return(0);
			}
			*p_edges=edges;
		break;
		case 'e':
	/*Note we do not bounds check _here_, the effect is accomplished by the
termination condition of the loop*/
			fscanf(in,
#ifdef BITS16
" %hu %hu\n"
#else
#ifdef BITS32
" %lu %lu\n"
#else
" %llu %llu\n"
#endif
#endif
, &edges[edgesfound].b, &edges[edgesfound].e);

			edges[edgesfound+1].b=edges[edgesfound].e;
			edges[edgesfound+1].e=edges[edgesfound].b;

	/*This is technically legal, but only confuses things so we ignore it
and print a warning*/
			if(edges[edgesfound].b==edges[edgesfound].e) {
				fprintf(stderr, "Warning: Edge from node %lld "
"to itself\n", (uint64)(edges[edgesfound].b));
				fprintf(stdout, "c Warning: Edge from node "
"%lld to itself\n", (uint64)(edges[edgesfound].b));
				--ecnt;
				break;
			}

	/*This is definitly illegal, however we can recover so we do; this may
not be the right decision, because time may be wasted working on a wrong
problem*/
			if(!edges[edgesfound].b||!edges[edgesfound].e||
edges[edgesfound].b>ecnt||edges[edgesfound].e>ecnt) {
				fprintf(stderr, "Warning: Illegal edge from "
"%lld to %lld\n", (uint64)(edges[edgesfound].b), (uint64)(edges[edgesfound].e));
				fprintf(stdout, "c Warning: Illegal edge from "
"%lld to %lld\n", (uint64)(edges[edgesfound].b), (uint64)(edges[edgesfound].e));
				--ecnt;
				break;
			}
			edgesfound+=2;
		break;
		case EOF:
	/*Again illegal, but recoverable*/
			fprintf(stderr, "Warning: Premature termination of "
"input\n");
			fprintf(stdout, "c Warning: Premature termination of "
"input\n");
			ecnt=edgesfound>>1;
		break;
		default:
			if(ltype&~0xFF) fprintf(stderr, "Read error #%d\n",
ltype);
			else fprintf(stderr, "Unexpected command: '%c'\n",
ltype);
			free(edges);
			return(0);
		break;
		}
	} while(edgesfound>>1<ecnt);

	do ltype=getc(in); while(isspace(ltype)||ltype=='\n');
	if(ltype!=EOF) {
		fprintf(stderr, "Warning: Input not fully consumed\n");
		fprintf(stdout, "c Warning: Input not fully consumed\n");
	}
	/*A bit of protection on the end*/
	edges[edgesfound].e=edges[edgesfound].b=0;


	if(!(ret=malloc(sizeof(struct instance)))) {
		fprintf(stderr, "Error: Unable to allocate instance "
"structure\n");
		free(edges);
		return(0);
	}
	*p_ret=ret;

	ret->islands=NULL;
	ret->nodecnt=nodes-1;
	/*ret->nodeptr*/
	ret->workchain=NULL;
	ret->bestcolors=0;

/************************************************************************
*WARNING!!!  WARNING!!!  WARNING!!!  WARNING!!!  WARNING!!!  WARNING!!! *
*************************************************************************
*The below does not take into account alignment concerns. We can get	*
*away with this because struct node * is highly likely to be a multiple	*
*of the required alignment. I can see this potencially causing problems	*
*on strange or future architectures.					*
************************************************************************/
	/*Due to the stupid ANSI requirement for arrays to be at least one
element we've already got 'nodes' edge pointers; also note that nodes has
the node with the greatest number, not the count*/
	ret->nodeptr=malloc(sizeof(struct node)*(nodes+1)+
sizeof(struct node *)*(ecnt*2-nodes-1));
	if(!ret->nodeptr) {
		fprintf(stderr, "Error: Unable to allocate node table\n");
		free(edges);
		free(ret);
		return(0);
	}

	/*We've done all the allocation and prep, now it is time for digestion*/

	/*BTW in the first version we returned an instance, which is why ret is
named ret, as opposed to ecnt being named ret*/
	return(ecnt);
}


int processinput(struct instance *inst, struct edge *edges, uint64 ecnt)
{
	node_t x, y, z;
	node_t m;
	struct edge *edgewk;
	struct node *nodewk, *nodetmp, **pnodewk;

/*First step, organized the edges					*/
	qsort(edges, ecnt<<1, sizeof(struct edge), (int (*)(const void *,
const void *))edgecmp);

/*Second step, break the huge node block apart, and organize it for	*
*subsequent processing							*/
	edgewk=edges;
	nodewk=inst->nodeptr;
	z=0; /*Collect some data while we're at it*/
	for(x=1; x<=inst->nodecnt+1; ++x) {
		y=0;
		while(edgewk->b==x) {
/*Deliberatly above we setup a spare on the end of edgewk*/
			if(edgewk[0].b!=edgewk[1].b||edgewk[0].e!=edgewk[1].e)
++y;
			else if(edgewk->b<edgewk->e) {
	/*every edge is recorded twice so we make sure to only print one*/
				fprintf(stderr, "Warning: Duplicate edge from "
"%lld to %lld\n", (uint64)edgewk->b, (uint64)edgewk->e);
				fprintf(stderr, "c Warning: Duplicate edge "
"from %lld to %lld\n", (uint64)edgewk->b, (uint64)edgewk->e);
			}
			++edgewk;
		}

		/*initialize the node*/
		nodewk->id=x;
		nodewk->iptr=NULL;
		nodewk->trivial=0;
		nodewk->bestcolor=0;
		nodewk->color=1; /*marked for later in digest*/
		nodewk->diff=0;
		nodewk->disc=0;
		nodewk->colset=NULL;
		nodewk->ecnt=y;

		if(y>z) z=y;

		/*Turn m into a mask*/
		m=NODEMAX^(NODEMAX>>1);
		while(m>x) m>>=1;

		pnodewk=&inst->nodeptr;
		m>>=1;
		while(m) {
			if(m&x) pnodewk=&(*pnodewk)->sptr;
			else pnodewk=&(*pnodewk)->iptr;
			m>>=1;
		}
		*pnodewk=nodewk; /*Note that for x==1, we'll rewrite the top
pointer, but thats okay since we're at where it should be pointing anyway*/

		/*Advance*/
		nodewk=(struct node *)(nodewk->edges+y);
	}
	/*Setup the end marker*/
	*nodewk=(struct node){0, NULL, NULL, 1, 0, 0, 0, 0, NULL, 0};

	fprintf(stdout, "c Largest number of connected edges: %llu\nc\n",
(uint64)z);

/*At this point the nodes is organized as a binomal tree using iptr,	*
*and sptr the reason is because the nodes are variably sized and so we	*
*cannot	directly access any particular node simply via inst->nodeptr.	*/

/*Time to actually setup the edge pointers				*/
	edgewk=edges;
	nodewk=inst->nodeptr;
	for(x=1; x<=inst->nodecnt+1; ++x) {
		y=0;
		while(edgewk->b==x) {
			if(edgewk[0].b==edgewk[1].b&&edgewk[0].e==edgewk[1].e) {
				++edgewk;  /*dup*/
				continue;
			}

			z=edgewk->e;
/*Because inst->nodeptr is *not* an array we cannot directly find the target*/
			m=NODEMAX^(NODEMAX>>1);
			while(m>z) m>>=1;

			nodetmp=inst->nodeptr;
			m>>=1;
			while(m) {
				if(m&z) nodetmp=nodetmp->sptr;
				else nodetmp=nodetmp->iptr;
				m>>=1;
			}
#ifdef DEBUG
			if(nodetmp->id!=z) {
				fprintf(stderr, "Internal error #%08X:%08X\n",
__LINE__, __LINE__);
				exit(65);
			}
#endif

			nodewk->edges[y]=nodetmp;

			++y;
			++edgewk;
		}

#ifdef DEBUG
		if(y!=nodewk->ecnt) {
			fprintf(stderr, "Internal error #%08X:%08X\n",
__LINE__, __LINE__);
			exit(65);
		}
#endif

	/*Since this is advancing node by node, we can use this faster method*/
		nodewk=(struct node *)(nodewk->edges+nodewk->ecnt);
	}

	return(1);
}

int edgecmp(const struct edge *a, const struct edge *b)
{
/*	int ret;
	ret=a->b-b->b;
	if(!ret) ret=a->e-b->e;
	return(ret);*/
	/*In theory the difference could be greater than 1<<30 and so we
must directly test, rather than simply using subtraction*/
	return(a->b>b->b?1:a->b<b->b?-1:a->e>b->e?1:a->e<b->e?-1:0);
}


int digestinput(struct instance *inst)
{
	node_t x;
	struct node *nodewk, *nodetmp, *nodetmp2;
	struct instance *instwk;
	uint64 st0, st1, st2; /*statistics temporaries*/

/************************************************************************
*Edge arrays are now setup, time to start major digestion		*
************************************************************************/

/*First pass, filter out trivial nodes					*/

	st2=st1=st0=0;
	nodewk=inst->nodeptr;
	while(nodewk->id) {
		if(nodewk->ecnt<=1&&!nodewk->trivial) {
		/*Leaf, trivial, look for other trivial nodes*/
			nodetmp=nodewk;
			if(nodewk->ecnt) nodetmp2=nodewk->edges[0];
			else nodetmp2=NULL;

			do { /*Look for alternate trivial nodes*/
/*Notes: We don't worry about returning to where we were just at because that
*is marked trivial, and avoided. We're looking for nodes that have a single
*link back to the main "island", and all other attached nodes are trivial. If
*there are multiple trivial attached nodes it will be grabbed when the final
*one is marked. We also don't worry about nodetmp2 being a trivial node,
*because if it was why didn't we already get grabbed?*/
				nodetmp->trivial=1;
				nodetmp->color=0; /*clear the mark*/
				nodetmp->iptr=inst->workchain;
				nodetmp->sptr=nodetmp2; /*mark color depend*/
				inst->workchain=nodetmp;

				++st0; ++st1; ++st2;

				if(!(nodetmp=nodetmp2)) break;
#ifdef DEBUG
				if(nodetmp2->trivial) {
					fprintf(stderr, "Internal error "
"#%08X:%08X\n", __LINE__, __LINE__);
					exit(65);
				}
#endif
				nodetmp2=NULL;
				for(x=0; x<nodetmp->ecnt; ++x) {
					if(!nodetmp->edges[x]->trivial) {
						if(nodetmp2) {
							nodetmp=NULL;
							break;
						} else {
							nodetmp2=
nodetmp->edges[x];
							--st0; /*not bsc trv*/
							--st2; /*Not Island*/
						}
					}
				}
			} while(nodetmp);

			nodewk=inst->workchain;
		}
		nodewk=(struct node *)(nodewk->edges+nodewk->ecnt);
	}
	fprintf(stdout, "c %llu trivial nodes found, %llu basic, %llu trivial "
"islands\nc\n", st1, st0, st2);

/*Second pass, island hopping time					*/

	nodewk=inst->nodeptr;
	st2=st1=st0=0;
	while(nodewk->id) {
		if(nodewk->color) {
			++st0;
			if(!(instwk=malloc(sizeof(struct instance)))) {
				fprintf(stderr, "Error: Unable to allocate "
"island header\n");
				free(inst->nodeptr);
				do {
					instwk=inst->islands;
					free(inst);
				} while((inst=instwk));
				return(0);
			}
			instwk->islands=inst->islands;
			inst->islands=instwk;
			instwk->nodeptr=nodewk;
			instwk->workchain=NULL;
			instwk->bestcolors=0;
			st2+=instwk->nodecnt=colonize_island(NULL, nodewk);
			if(instwk->nodecnt>st1) st1=instwk->nodecnt;
		}
		nodewk=(struct node *)(nodewk->edges+nodewk->ecnt);
	}

	if(st0) fprintf(stdout, "c %llu islands found, largest is %llu nodes, "
"totalling %llu nodes\nc\n", st0, st1, st2);
	else fprintf(stdout, "c Instance identified as trivial\n");

	return(1);
}

node_t colonize_island(struct node *chain, struct node *n)
{
	node_t cnt=1;
	edge_t e=0;

#ifdef DEBUG
	if(!n->color) {
		fprintf(stderr, "Internal Error #%08X:%08X\n", __LINE__,
__LINE__);
		exit(65);
	}
#endif
	n->color=0;

	while(e<n->ecnt) {
		if(n->edges[e]->color) {
			cnt+=colonize_island(chain, n->edges[e]);
			chain=n->edges[e];
		}
		if(!n->edges[e]->trivial) ++n->diff; /*How annoying are we?*/
		++e;
	}
	n->iptr=chain;
	return(cnt);
}

