/* **********************************************************************
*Copyright (C) 2002 Elliott Mitchell					*
*All rights reserved.							*
*									*
*$Id: search.c,v 3.20 2002/06/03 05:35:32 ehem Exp $			*
*************************************************************************
*CS410-Combinatorial Search Project					*
************************************************************************/


#define DBGFILE 4

#include "common.h"
#include "libcolorset.h"


#ifndef MAXTOUCH
#define MAXTOUCH 1000
#endif

#ifndef ADJ
#ifndef NOADJ
#define ADJ
#endif
#endif


static int ckisord(struct instance *);


int search(struct instance *inst)
{
	struct instance *is;
	struct node *wk, *save;
	color_t col=0, maxcol; /*col need not be init, but avoids a warning*/
	edge_t e;
#ifdef ADJ
	struct node **base;
	edge_t top;
#endif
	uint64 touch=0;
#ifdef DEBUG
	uint64 stats0=0, stats1=0, stats2=0;
#endif

/************************************************************************
*This of course would be easier to do recursively, unfortunately that	*
*doesn't work. We may potencially be dealing with millions, perhaps	*
*hundreds of millions or even a couple billion nodes. It is nice to use	*
*lots of local variables, but these add up very quickly and our stack	*
*would quickly explode; more significantly the overhead of function	*
*calls would end up very large. Things are even worse than this though,	*
*Linux imposes a maximum limit of 8MB on the stack (other OSes may	*
*impose other limits, although those could be a lot higher), more than	*
*anything else this underscores the point that by the time you're	*
*dealing with that much data it should be on the heap. Assuming each	*
*frame was 16 bytes (4 for SP, and 3 variables) we're already limited	*
*to merely 524,288 nodes; so we avoid the stack.			*
************************************************************************/

	while((is=inst->islands)&&touch<=MAXTOUCH) {
		maxcol=is->bestcolors;
		wk=is->workchain; /*This is set to magically work in dumb.c*/

#ifndef TOUCHNODES
		++touch;
#endif

#ifdef ADJ
/*Notes on this test: If the node's color is >= maxcol then this is /a/ node
that exploded maxcol, and so must be delt with; the only way to deal with it is
to decrement a neighbor. This node might of exploded due to increasing it's
color to help somebody else, if so then another neighbor of that problem node
could solve the problem while letting this one return would solve the
problem.*/
		if(wk->color>=maxcol&&wk->disc==wk->sptr->disc) save=wk;
		else save=NULL;
#endif

		do { /*reverse*/
#ifdef TOUCHNODES
			++touch;
#endif
			wk->color=0;
			wk=wk->sptr;
			if(!wk->sptr) { /*Hit the begining*/
				++wk->disc;
/*				wk->color=1;*/ /*The first can be constant*/
				fprintf(stdout, "c Increasing discrepancy "
"limit to %llu for an island of %llu nodes\nc\n", (unsigned long long)wk->disc,
(unsigned long long)is->nodecnt);
				break;
			}
			col=wk->color;
#ifdef DEBUG
			if(col>=maxcol) { /*This shouldn't happen*/
				fprintf(stderr, "Internal error "
"#%04X%04X:%08X\n", DBGFILE, __LINE__, __LINE__);
				continue;
			}
#endif

			/*Grab a discrepancy and try next*/
#ifndef ALTPENALTY
			if(!wk->disc) { col=maxcol; continue; }
#else
			if(wk->disc>=wk->sptr->disc) {
				if(!wk->disc) { col=maxcol; continue; }
				--wk->disc;
			}
#endif

#ifdef ADJ
			if(save) { /*Test for adjacency with problem node*/
/*Note that we explicitly look at save->edges, rather than wk->edges. The
difference is pretty small, except that save doesn't change. This means
save->edges will be loaded into processor cache *once*, while using wk->edges
would result in much larger areas of memory being loaded into processor cache
(the effect is greater with sparse graphs, lesser with dense graphs).*/
/*Also note that we assume the edges array is sorted (and in fact because the
way nodes are allocated in memory we can actually compare the /pointers/.*/

				base=save->edges;
				top=save->ecnt;

				do {
					e=top>>1;
					if(base[e]<wk) {
						base=base+e+1;
						top-=e+1;
					} else if(base[e]>wk) top=e;
					else {
						save=NULL;
						break;
					}
				} while(top); /*More efficient than while()*/

				if(save) {
					col=maxcol;
					continue;
				}
			}
#endif

#ifndef ALTPENALTY
			--wk->disc;
#endif
			do ++col; while(CS_TST(wk->colset, col));
		} while(col>=maxcol); /*...if possible*/

#ifdef DEBUG
		if(wk->disc) ++stats1;
		else ++stats0;
		if(wk->sptr&&wk->disc<wk->sptr->disc) ++stats2;
#endif

		/*Note we never actually touch the first node's color; this
gets set to 1 during dumb, however it is perfectly acceptable to leave it
statically set since one single node's color is irrelevant anyway*/

		do { /*forward*/
#ifdef TOUCHNODES
			++touch;
#endif
			wk=wk->iptr;
			if(wk->colset) CS_CLEAR(wk->colset);
			else CS_NEW(wk->colset);
			for(e=0; e<wk->ecnt; ++e) CS_SET(wk->colset,
wk->edges[e]->color);
			col=1;
			while(CS_TST(wk->colset, col)) ++col;

			wk->color=col;

			/*This needs to be done early to keep reverse code
consistant*/
			wk->disc=wk->sptr->disc; /*Find our limit*/

/*			if(col>=maxcol) break;*/ /*Aieeee! (BnB)*/
		} while(wk->iptr&&col<maxcol);
		is->workchain=wk; /*Save this useful bit of information*/
		if(col>=maxcol) continue; /*Other half of the Aieeee!, above*/

		save=wk; /*Save for stats later*/

		/*Save the improved solution*/
		maxcol=2; /*The smallest possible value*/
		do { /*We need to compute the new max, and save this*/
			if((col=wk->bestcolor=wk->color)>maxcol) {
				maxcol=col;
				/*Make sure we restart somewhere useful*/
				is->workchain=wk;
			}
			wk=wk->sptr;
		} while(wk);

#ifdef DEBUG
		if(maxcol>=is->bestcolors) {
			fprintf(stderr, "Internal error #%04X%04X:%08X\n",
DBGFILE, __LINE__, __LINE__);
			exit(65);
		}
#endif

		is->bestcolors=maxcol;
		if(is->islands) {
			if(is->islands->bestcolors>maxcol) { /*YAY!*/
				inst->bestcolors=is->islands->bestcolors;
				if(maxcol>2) ckisord(inst);
			} else inst->bestcolors=maxcol;
		} else inst->bestcolors=maxcol;

		fprintf(stdout, "c Found improved solution with %llu colors "
"for an island of %llu nodes\n" "c %llu discrepancies leftover; checked %llu "
"solutions, %schanging islands\nc\n", (unsigned long long)maxcol,
(unsigned long long)is->nodecnt, (unsigned long long)save->disc, touch,
inst->islands==is?"not ":"");

		if(maxcol<=2) { /*The holy grail*/
			fprintf(stdout, "c Sank an island of %llu nodes\n"
"c\n", (uint64)is->nodecnt);

			for(wk=is->nodeptr; wk; wk=wk->iptr) CS_DEL(wk->colset);
			inst->islands=is->islands;
			free(is);
		}

	}

#ifdef DEBUG
	fprintf(stdout, "c stats: %llu, %llu, %llu\nc\n", stats0, stats1,
stats2);
#endif

/*	trivialcolor(inst);*/
	return(1);
}

int ckisord(struct instance *targ)
{
	struct instance **lastp;
	struct instance *nxt;

	lastp=&targ->islands;
	targ=targ->islands;
	nxt=targ->islands;

	/*We preserve order by default. If it took a lot of work just to
get to equaling another node, it will likely be a lot more work to get below
than to push the other node down. Also produces favorable cache behavior
(yeah right) because we preserve locality*/
	while(nxt&&targ->bestcolors<nxt->bestcolors) {
		*lastp=nxt;
		lastp=&nxt->islands;
		nxt=nxt->islands;
	}
	targ->islands=nxt;
	*lastp=targ;
	return(1);
}

